__author__ = "Marc-Olivier Derouin"
__email__ = "marcolivier.derouin@poulet-frit.com"

import threading
import _thread
import unittest
from unittest.mock import patch
from abc import ABC, abstractmethod
from typing import List, Callable, Optional, Union
from io import StringIO
from collections import deque
import sys
import importlib.util
import os
import inspect
import argparse
from school_grader.html_test_result import HTMLTestResult
from school_grader.json_test_result import JSONTestResult
from school_grader.equality import Equality


def timeout(timeout_time: Optional[int]):
    """
    Decorator that sets a timer for a function execution.
    If the function execution takes longer than `timeout_time` seconds,
    the main thread of the program will be interrupted.

    Args:
        timeout_time (int or float): The number of seconds after which to interrupt the main thread.

    Returns:
        callable: The decorated function.
    """
    def decorator(function: Callable):
        """
        The actual decorator function that sets the timer and raises an exception
        if the function execution takes longer than `timeout_time` seconds.

        Args:
            function: The function to be decorated.

        Returns:
            The decorated function `inner`.
        """
        def inner(*args, **kwargs):
            """
            The decorated function that sets the timer and raises an exception
            if the function execution takes longer than `timeout_time` seconds.

            Args:
                *args: The positional arguments passed to the decorated function.
                **kwargs: The keyword arguments passed to the decorated function.
            """
            timer = None
            try:
                if timeout_time is not None:
                    timer = threading.Timer(timeout_time, lambda: _thread.interrupt_main())
                    timer.start()
                function(*args, **kwargs)
            except KeyboardInterrupt:
                raise TimeoutError(f'Program execution did not finish within the allotted time of {timeout_time} seconds. You may be in an infinite loop.')
            finally:
                if timer is not None:
                    timer.cancel()
        return inner
    return decorator


class TestCase(unittest.TestCase, ABC):
    """Abstract base class for test cases."""
    tests_case = []

    def __init__(self, test_name: str, timeout_time: float, fail_message: str):
        super().__init__()
        self.tests_case.append(self)
        self.shortDescription = lambda: test_name
        self._timeout_time = timeout_time
        self._fail_message = fail_message
        self._dirname = os.getcwd()
        self.line_number = inspect.currentframe().f_back.f_back.f_lineno

    @abstractmethod
    def runTest(self) -> None:
        """Run the test case."""
        pass


class FileTestCase(TestCase):
    """Custom test case class for testing python code"""

    def __init__(self, test_name: str, file_name: str, expected_output: List[Union[str, Equality]], mock_input: List[str] = [], timeout_time: float = 1, fail_message: str = None):
        """
        Initializes the test case with the given attributes.

        Args:
            test_name (str): Name of the test.
            file_name (str): Name of the Python file to be tested.
            expected_output (list[str] or list[Equality]): List of str or Equality objects to be used as expected outputs.
            mock_input (list[str]): List of strings to be used as inputs.
            timeout_time (int or float): Maximum time in seconds that the test can run.
            fail_message (str): Message to be displayed in case of failure.

        Returns:
            None
        """
        super().__init__(test_name, timeout_time, fail_message)
        self._file_name = file_name
        self._mock_input = mock_input
        self._expected_output = expected_output

    def run_whole_file(self) -> None:
        """
        Run the whole file.

        Returns:
            None
        """
        try:
            spec = importlib.util.spec_from_file_location(self._file_name, os.path.join(self._dirname, f'{self._file_name}.py'))
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        except FileNotFoundError as e:
            raise FileNotFoundError(f'File {self._file_name}.py not found. Your current working directory is {self._dirname}.') from e

    def override_input(self) -> Callable[[], str]:
        """
        Override the input function to use the mock input.

        Returns:
            A function that returns the next mock input.
        """
        queue = deque(self._mock_input)
        def fake_input(*_, **__) -> str:
            if not queue:
                raise AssertionError("Too many input calls. Check your code")
            return queue.popleft()
        return fake_input

    def runTest(self):
        """
        Run the test case.

        Returns:
            None
        """
        @timeout(self._timeout_time)
        def exec_test():
            """
            Execute the test case.

            Returns:
                None
            """
            with patch('builtins.input', self.override_input()), patch('sys.stdout', new=StringIO()) as fake_out:
                try:
                    self.run_whole_file()
                    output = fake_out.getvalue().strip().splitlines()

                    if (len(output) != len(self._expected_output)):
                        raise AssertionError(f'The output of your program contains {len(output)} lines. You should have {len(self._expected_output)} lines')
                    for result, expected in zip(output, self._expected_output):
                        if isinstance(expected, str):
                            self.assertEqual(result, expected,self._fail_message)
                        elif isinstance(expected, Equality):
                            expected.validate(self, result, self._fail_message)
                        else:
                            raise TypeError(f'Expected output should be a string or an Equality object. Got {type(expected)} instead.')
                finally:
                    sys.stdout = sys.__stdout__
        exec_test()


class FunctionTestCase(TestCase):
    """A basic test case class for running unit tests."""

    def __init__(self, test_name: str, result_func: Callable, expected_result, timeout_time: float = 1, fail_message: str = None):
        """
        Initialize the test case with the required attributes.

        Args:
            test_name (str): The name of the test case.
            result_func (function): A function that returns the result to be tested.
            expected_result: The expected result from the result_func.
            timeout_time (int or float): The maximum time in seconds allowed for the test to complete.
            fail_message (str, optional): A custom message to display on test failure. Default is None.

        Returns:
            None
        """
        super().__init__(test_name, timeout_time, fail_message)
        self._result_func = result_func
        self._expected_result = expected_result

    def runTest(self):
        """Execute the test case and check the result against the expected result."""
        @timeout(self._timeout_time)
        def exec_test():
            """Execute the test within a time limit defined by the timeout attribute."""
            self.assertEqual(self._expected_result, self._result_func(), self._fail_message)
        exec_test()


def run_tests(generate_html: bool = True):
    """
    Run all the tests in the BasicTestCase and FunctionTestCase classes.

    Args:
        generate_html (bool): Whether to generate an HTML report of the test results. Default is True.

    Returns:
        bool: The value of the `generate_html` parameter indicating whether an HTML report is generated.
    """
    if not TestCase.tests_case:
        raise ValueError("No test cases were added to the test suite.")
    
    parser = argparse.ArgumentParser()
    parser.add_argument('--extension', action='store_true')
    args = parser.parse_args()

    suite = unittest.TestSuite()
    suite.addTests(TestCase.tests_case)

    def json_runner() -> None:
        """
        Run the test suite and generate a JSON report of the results.

        Args:
            None

        Returns:
            None
        """
        with patch('sys.stdout', new=StringIO()) as _:
            runner = unittest.TextTestRunner(resultclass=JSONTestResult)
            result: JSONTestResult = runner.run(suite)
        result.print_report()

    def html_runner() -> None:
        """
        Run the test suite and generate an HTML report of the results.

        Args:
            None

        Returns:
            None
        """
        runner = unittest.TextTestRunner(resultclass=HTMLTestResult)
        result: HTMLTestResult = runner.run(suite)
        result.generate()

    def text_runner() -> None:
        """
        Run the test suite and generate a text report of the results.

        Args:
            None

        Returns:
            None
        """
        runner = unittest.TextTestRunner()
        runner.run(suite)
    
    if args.extension:
        json_runner()
        return

    if generate_html:
        html_runner()
    else:
        text_runner()



__author__ = "Marc-Olivier Derouin"
__email__ = "marcolivier.derouin@poulet-frit.com"

from school_grader.html_test_result import HTMLTestResult
from school_grader.json_test_result import JSONTestResult
from school_grader.test_case import FileTestCase, FunctionTestCase, run_tests, TestCase, timeout
from school_grader.equality import Equality, AlmostEqualNumber, AlmostEqualString, CaseInsensitiveStringEquality, WhiteSpaceInsensitiveEquality, ContainsEquality, CombineEqualities



__author__ = "Marc-Olivier Derouin"
__email__ = "marcolivier.derouin@poulet-frit.com"

import re
from typing import List, Tuple
import unittest
from abc import ABC, abstractmethod


class Equality(ABC):
    """Abstract base class for equality definitions"""

    def __init__(self, expected: str):
        """Initialize the class with the expected value.

        Args:
            expected: The expected value as a string.
        """
        self.expected = expected

    @abstractmethod
    def validate(self, test_case: unittest.TestCase, value_to_test: str, fail_message: str) -> None:
        """Validate equality between `value_to_test` and the expected value.

        Args:
            test_case: Instance of `unittest.TestCase` to run the equality check.
            value_to_test: The value to test for equality.
            fail_message: The message to display if the equality check fails.
        """
        pass


class AlmostEqual(Equality):
    """Abstract base class for almost equal validation."""
    pass


class AlmostEqualString(AlmostEqual):
    """Class for almost equal string validation."""

    def __init__(self, expected: str, max_distance: int = 2):
        """Initialize the class with the expected value and the Levenshtein distance.

        Args:
            expected: The expected value as a string.
            distance: The Levenshtein distance. Defaults to 2.
        """
        super().__init__(expected)
        self.max_distance = max_distance

    @staticmethod
    def _levenshtein(s1: str, s2: str) -> int:
        """Calculate the Levenshtein distance between two strings.

        Args:
            s1: The first string.
            s2: The second string.

        Returns:
            The Levenshtein distance between `s1` and `s2`.
        """
        # Create a matrix with dimensions (length of s1 + 1) x (length of s2 + 1)
        matrix = [[0] * (len(s2) + 1) for _ in range(len(s1) + 1)]

        # Initialize the first row and column of the matrix
        for i in range(len(s1) + 1):
            matrix[i][0] = i
        for j in range(len(s2) + 1):
            matrix[0][j] = j

        # Fill in the rest of the matrix
        for i in range(1, len(s1) + 1):
            for j in range(1, len(s2) + 1):
                cost = 0 if s1[i - 1] == s2[j - 1] else 1
                matrix[i][j] = min(
                    matrix[i - 1][j] + 1,           # deletion
                    matrix[i][j - 1] + 1,           # insertion
                    matrix[i - 1][j - 1] + cost     # substitution
                )

        # The final value in the bottom-right corner of the matrix is the Levenshtein distance
        return matrix[-1][-1]

    def validate(self, test_case: unittest.TestCase, value_to_test: str, fail_message: str) -> None:
        """Validate almost equal string equality between `value_to_test` and the expected value.

        Args:
            test_case: Instance of `unittest.TestCase` to run the equality check.
            value_to_test: The value to test for equality.
            fail_message: The message to display if the equality check fails.
        """
        test_case.assertLessEqual(AlmostEqualString._levenshtein(value_to_test, self.expected), self.max_distance, fail_message)


class AlmostEqualNumber(AlmostEqual):
    """Class for almost equal numerical validation."""

    def __init__(self, expected: str, precision: int = 7):
        """Initialize the class with the expected value and the precision values.

        Args:
            expected: The expected value as a string.
            precisions: The number of decimal places to compare. Defaults to 7.
        """
        super().__init__(expected)
        self.precision = precision

    def validate(self, test_case: unittest.TestCase, value_to_test: str, fail_message: str) -> None:
        """Validate almost equal numerical equality between `value_to_test` and the expected value.

        Args:
            test_case: Instance of `unittest.TestCase` to run the equality check.
            value_to_test: The value to test for equality.
            fail_message: The message to display if the equality check fails.
        """
        expected_numbers = list(map(float, re.findall(r'[-+]?\d*\.\d+|\d+', self.expected)))
        result_numbers = list(map(float, re.findall(r'[-+]?\d*\.\d+|\d+', value_to_test)))

        if len(expected_numbers) != len(result_numbers):
            raise AssertionError(f"Expected {len(expected_numbers)} numbers, but got {len(result_numbers)}")

        for expected_number, result_number in zip(expected_numbers, result_numbers):
            test_case.assertAlmostEqual(float(expected_number), float(result_number), self.precision, fail_message)


class Equal(Equality):
    """Abstract base class for equality validation."""
    @staticmethod
    @abstractmethod
    def preprocess_for_comparison(expected: str, value_to_test: str) -> Tuple[str, str]:
        """Alter the expected and value_to_test values before running the equality check.

        Args:
            value_to_test: The value to test for equality.

        Returns:
            A tuple containing the expected and value_to_test values after alteration.
        """
        pass

    def validate(self, test_case: unittest.TestCase, value_to_test: str, fail_message: str) -> None:
        """Validate after altering .

        Args:
            test_case: Instance of `unittest.TestCase` to run the equality check.
            value_to_test: The value to test for equality.
            fail_message: The message to display if the equality check fails.
        """
        expected, value_to_test = self.preprocess_for_comparison(self.expected, value_to_test)
        test_case.assertEqual(expected, value_to_test, fail_message)

def CombineEqualities(*equalities: List[Equal]) -> Equal:
    """Combine multiple equalities into a single equality.

    Args:
        *equalities: The equalities to combine.

    Returns:
        A single equality that combines all of the provided equalities.
    """
    class CombinedEquality(Equal):
        """Class for combined equality validation."""

        def __init__(self, expected: str):
            """Initialize the class with the expected string value.

            Args:
                expected: The expected string value.
            """
            super().__init__(expected)

        @staticmethod
        def preprocess_for_comparison(expected: str, value_to_test: str) -> Tuple[str, str]:
            """Alter the expected and value_to_test values before running the equality check.

            Args:
                value_to_test: The value to test for equality.

            Returns:
                A tuple containing the expected and value_to_test values after alteration.
            """
            for equality in equalities:
                if not isinstance(equality, Equal):
                    raise TypeError(f"Expected an instance of Equal, but got {type(equality)}")
                expected, value_to_test = equality.preprocess_for_comparison(expected, value_to_test)
            return expected, value_to_test
    return CombinedEquality


class CaseInsensitiveStringEquality(Equal):
    """Class for case-insensitive string equality validation."""

    def __init__(self, expected: str):
        """Initialize the class with the expected string value.

        Args:
            expected: The expected string value.
        """
        super().__init__(expected)

    @staticmethod
    def preprocess_for_comparison(expected: str, value_to_test: str) -> Tuple[str, str]:
        """Alter the expected and value_to_test values before running the equality check by converting both to lowercase.

        Args:
            expected: The expected value.
            value_to_test: The value to test for equality.

        Returns:
            A tuple containing the expected and value_to_test values after alteration by converting both to lowercase.
        """
        return expected.lower(), value_to_test.lower()


class WhiteSpaceInsensitiveEquality(Equal):
    """Class for space-insensitive equality validation."""

    def __init__(self, expected: str):
        """Initialize the class with the expected string value.

        Args:
            expected: The expected string value.
        """
        super().__init__(expected)

    @staticmethod
    def preprocess_for_comparison(expected: str, value_to_test: str) -> Tuple[str, str]:
        """Alter the expected and value_to_test values before running the equality check by removing all whitespace.

        Args:
            expected: The expected value.
            value_to_test: The value to test for equality.

        Returns:
            A tuple containing the expected and value_to_test values after alteration by removing all whitespace.
        """
        return re.sub(r"\s+", "", expected), re.sub(r"\s+", "", value_to_test)


class ContainsEquality(Equal):
    """Class for contains equality validation. Checks if the expected value is contained in the value to test."""

    def __init__(self, expected: str):
        """Initialize the class with the expected string value.

        Args:
            expected: The expected string value. This value will be checked to see if it is contained in the value to test.
        """
        super().__init__(expected)

    @staticmethod
    def _find_largest_substring(word: str, target: str) -> str:
        word_length = len(word)
        target_length = len(target)

        # Create a table to store the lengths of common substrings
        table: List[List[int]] = [[0] * (target_length + 1) for _ in range(word_length + 1)]

        # Variables to track the longest common substring and its length
        max_length = 0
        end_index = 0

        # Fill the table using dynamic programming
        for i in range(1, word_length + 1):
            for j in range(1, target_length + 1):
                if word[i - 1] == target[j - 1]:
                    table[i][j] = table[i - 1][j - 1] + 1
                    if table[i][j] > max_length:
                        max_length = table[i][j]
                        end_index = i

        # Extract the longest common substring
        largest_substring = word[end_index - max_length:end_index]
        return largest_substring

    @staticmethod
    def preprocess_for_comparison(expected: str, value_to_test: str) -> Tuple[str, str]:
        """Alter the expected and value_to_test values before running the equality check by keeping only the expected value if it is contained in the value to test.

        Args:
            expected: The expected value.
            value_to_test: The value to test for equality.

        Returns:
            A tuple containing the expected and value_to_test values after alteration by keeping only the expected value if it is contained in the value to test.
        """
        if expected in value_to_test:
            return expected, expected
        else:
            # Return the biggest substring of the value to test that is contained in the expected value.
            return expected, ContainsEquality._find_largest_substring(value_to_test, expected)



__author__ = "Marc-Olivier Derouin"
__email__ = "marcolivier.derouin@poulet-frit.com"

import unittest
import json
from dataclasses import dataclass

@dataclass
class Result:
    """A class for the result of a test."""
    description: int
    status: str
    message: str    


class JSONTestResult(unittest.TextTestResult):
    """A class for the JSON result of a test."""

    def __init__(self, stream, descriptions, verbosity):
        super().__init__(stream, descriptions, verbosity)
        self._test_result = dict()

    def addSuccess(self, test):
        """Adds a successful test to the JSON result."""
        self._test_result[test.line_number] = Result(test.shortDescription(), "success", "Test passed").__dict__

    def addFailure(self, test, err):
        """Adds a failed test to the JSON result."""
        self._test_result[test.line_number] = Result(test.shortDescription(), "failure", self._exc_info_to_string(err, test)).__dict__
        

    def addError(self, test, err):
        """Adds an error test to the JSON result."""
        self._test_result[test.line_number] = Result(test.shortDescription(), "error", self._exc_info_to_string(err, test)).__dict__

    def print_report(self):
        """Prints the JSON result of a test."""
        print(json.dumps(self._test_result))
        



__author__ = "Marc-Olivier Derouin"
__email__ = "marcolivier.derouin@poulet-frit.com"

from abc import ABC, abstractmethod
from typing import List
import unittest
import webbrowser


HTML_BEGIN = (
    "<html>"
    "<head>"
    '<link href="https://fonts.googleapis.com/css2?family=Roboto:wght@400;500;700&display=swap" rel="stylesheet">'
    "<style>"
    "table {"
    "border-radius: 10px;"
    "width: 100%;"
    "margin: 40px auto;"
    "box-shadow: 0 2px 15px rgba(0, 0, 0, 0.1);"
    "text-align: center;"
    "border-spacing: 0;"
    "}"
    "table tr:first-child td {"
    "border-top: 0;"
    "}"
    "table tr td:first-child {"
    "border-left: 0;"
    "}"
    "table tr:last-child td {"
    "border-bottom: 0;"
    "}"
    "table tr td:last-child {"
    "border-right: 0;"
    "}"
    "th, td {"
    "border: solid 1px #68706a29;"
    "padding: 20px;"
    "font-size: 16px;"
    "font-weight: 500;"
    "color: #333;"
    "text-align: center;"
    "vertical-align: middle;"
    "}"
    "th {"
    "background-color: #ddd;"
    "font-weight: bold;"
    "}"
    "pre {"
    "background-color: #f6f8fa;"
    "border-radius: 10px;"
    "font-size: 85%;"
    "line-height: 1.45;"
    "overflow: auto;"
    "padding: 16px;"
    "text-align: left;"
    "margin: 20px 0;"
    "}"
    "body {"
    "padding: 20px;"
    "background-color: #f2f2f2;"
    "font-family: 'Roboto', sans-serif;"
    "}"
    ".error {"
    "background-color: #FFE844;"
    "}"
    ".pass {"
    "background-color: #A6FB88;"
    "}"
    ".fail {"
    "background-color: #FF4444;"
    "}"
    "</style>"
    "</head>"
    "<body>"
    "<table>"
    "<tr>"
    "<th>Test name</th>"
    "<th>Directory</th>"
    "<th>Result</th>"
    "<th>Stack trace</th>"
    "</tr>"
)

HTML_END = (
    "</table>"
    "</body>"
    "</html>"
)


def sanitize_html(html_string: str) -> str:
    return html_string.replace("<", "&lt;").replace(">", "&gt;")


class HTMLResult(ABC):
    """An abstract class for the HTML result of a test."""

    def __init__(self, test):
        self._test = test

    @abstractmethod
    def generate_html(self) -> str:
        """Generates the HTML result of a test."""
        return (
            "<tr>"
            f"<td><b>{sanitize_html(self._test.shortDescription())}</b></td>"
            f"<td><b>{sanitize_html(self._test._dirname)}</b></td>"
        )

class HTMLSuccessResult(HTMLResult):
    """A class for the HTML result of a successful test."""

    def __init__(self, test):
        super().__init__(test)

    def generate_html(self) -> str:
        """Generates the HTML result of a successful test."""
        return (
            super().generate_html() +
            "<td class=\"pass\"><b>PASS</b></td>"
            "<td></td>"
            "</tr>"
        )


class HTMLFailureResult(HTMLResult):
    """A class for the HTML result of a failed test."""

    def __init__(self, test, stack_trace: str):
        super().__init__(test)
        self._stack_trace = stack_trace

    def generate_html(self) -> str:
        """Generates the HTML result of a failed test."""
        return (
            super().generate_html() +
            "<td class=\"fail\"><b>FAIL</b></td>"
            f"<td><pre>{sanitize_html(self._stack_trace)}</pre></td>"
            "</tr>"
        )


class HTMLErrorResult(HTMLResult):
    """A class for the HTML result of a failed test."""

    def __init__(self, test, stack_trace: str):
        super().__init__(test)
        self._stack_trace = stack_trace

    def generate_html(self) -> str:
        """Generates the HTML result of a failed test."""
        return f"""
        {super().generate_html()}
            <td class="error"><b>ERROR</b></td>
            <td><pre>{sanitize_html(self._stack_trace)}</pre></td>
        </tr>
        """


class HTMLTestResult(unittest.TextTestResult):
    """A class for the HTML result of a test."""

    def __init__(self, stream, descriptions, verbosity):
        super().__init__(stream, descriptions, verbosity)
        self._test_result: List[HTMLResult] = []

    def addSuccess(self, test):
        """Adds a successful test to the HTML result."""
        self._test_result.append(HTMLSuccessResult(test))

    def addFailure(self, test, err):
        """Adds a failed test to the HTML result."""
        self._test_result.append(HTMLFailureResult(test, self._exc_info_to_string(err, test)))

    def addError(self, test, err):
        """Adds an error test to the HTML result."""
        self._test_result.append(HTMLErrorResult(test, self._exc_info_to_string(err, test)))

    def generate(self):
        """Generates the HTML result of a test."""
        html_content = HTML_BEGIN
        for result in self._test_result:
            html_content += result.generate_html()
        html_content += HTML_END
        with open("results.html", "w") as f:
            f.write(html_content)
        webbrowser.open("results.html", new=2)